--============================================================================--
--------------------------------------------------------------------------------
--  PALETTE.E Palette routines for the Euphoria DOS Interpretor.
--  Version 1.0 alpha (c)2004. Hayden McKay. FREEWARE Use Freely :).

--  This library is free software; you can redistribute it and/or
--  modify it under the terms of the GNU Lesser General Public
--  License as published by the Free Software Foundation; either
--  version 2.1 of the License, or (at your option) any later version.

--  This library is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
--  Lesser General Public License for more details.

--  You should have received a copy of the GNU Lesser General Public
--  License along with this library; if not, write to the Free Software
--  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--------------------------------------------------------------------------------

------------------------------[****IMPORTANT****]-------------------------------
--                                                                            --
-- This module uses direct RAMDAC access to retrieve/update the VGA palette.  --
-- This meens that SuperVGA modes are most likley to fail. On many if not all --
-- SuperVGA/XGA adaptors the standard VGA registers no longer exsist except   --
-- in standard EGA/VGA modes. All standard EGA/VGA modes wont be affected.    --
-- In laymens terms PALETTE.E is compatable with modes supplied by Euphoria.  --
--                                                                            --
--------------------------------------------------------------------------------

-- Comments/Bugs/Suggestions => hayden_mckay@hotmail.com

-- Please Post or E-Mail me if direct RAMDAC access fails on your adaptor
-- when in Euphoria graphics modes. Pls leave me the resolution details.
-- * n.b I may have to devise a NonVGA check for VESA modes.

-- Notes:   1.  There is no error detection or type checking.

--          2.  This version is 'alpha' so Mail your bugs please.

--          3.  These routines will also except 'Direct Color' palettes.
--              eg; A palette/palette-array that consists of {R, G, B, X}.
--              Remember to scale them down before applying any routine.

--------------------------------------------------------------------------------
--============================================================================--
--------------------------------------------------------------------------------
include machine.e
include misc.e

constant stack = allocate(4)     --> stack to hold a maximum DWORD
lock_memory(stack, 4)            --> exclusive rights

--------------------------------------------------------------------------------

constant SetVGAPalette = allocate(22)
constant GetVGAPalette = allocate(32)
poke(GetVGAPalette, {
    #60,                    --    0: pushad
    #BD,#00,#00,#00,#00,    --    1: mov ebp stack (2)
    #66,#C7,#C2,#C7,#03,    --    6: mov dx, #03C7
    #B0,#00,                --    B: mov al, color (12)
    #EE,                    --    D: out dx, al
    #66,#83,#C2,#02,        --    E: add dx, 2
    #EC,                    --   12: in al, dx
    #88,#45,#00,            --   13: mov [ebp], al
    #EC,                    --   16: in al, dx
    #88,#45,#01,            --   17: mov [ebp+1], al
    #EC,                    --   1A: in al, dx
    #88,#45,#02,            --   1B: mov [ebp+2], al
    #61,                    --   1E: popad
    #C3})                   --   1F: ret
poke4(GetVGAPalette + 2, stack)
constant GetVGAPaletteCol = GetVGAPalette + 12

--------------------------------------------------------------------------------

-- sequence = GetAllPalette()

-- Gets the entire number of color palettes and returns an array representing
-- the current palette values.

-- Output:
-- Palette  :   {{R, G, B}, {R, G, B}, etc...}  :   256 x 3-bytes

global function GetAllPalette()
    sequence Palette
    
    Palette = {}
    for i = 0 to 255 do
	poke(GetVGAPaletteCol, i)
	call(GetVGAPalette)
	Palette = append(Palette, peek({stack, 3}))
    end for

    return Palette

end function
integer PeekAllPalettes PeekAllPalettes = routine_id("GetAllPalette")

--------------------------------------------------------------------------------

-- SetAllPalette(sequence palarray)

-- Sets the entire number of color palettes given an array of palette data.
-- * n.b. The palette array is poked directly to the RAMDAC.

-- Input:
-- Palette  :   {{R, G, B}, {R, G, B}, etc...}  :   256 x 3-bytes

global procedure SetAllPalette(sequence Palette)
    integer Color
    
    for i = 1 to 256 do
	Color = i - 1
	poke(SetVGAPalette,
	{#60,                   --    0: pushad
	#66,#BA,#C8,#03,        --    1: mov dx, #03C8
	#B0,Color,              --    5: mov al, color (6)
	#EE,                    --    7: out dx, al
	#66,#FF,#C2,            --    8: inc dx
	#B0,Palette[i][1],      --    B: mov al, red (12)
	#EE,                    --    D: out dx, al
	#B0,Palette[i][2],      --    E: mov al, green (15)
	#EE,                    --   10: out dx, al
	#B0,Palette[i][3],      --   11: mov al, blue (18)
	#EE,                    --   13: out dx, al
	#61,                    --   14: popad
	#C3}                    --   15: ret
	)
	call(SetVGAPalette)
    end for

end procedure
integer PokeAllPalettes PokeAllPalettes = routine_id("SetAllPalette")

--------------------------------------------------------------------------------

-- SetPalette(integer color, sequence palette)
-- Sets the desired color palette given a sequence of RGB palette data.
-- * n.b. The palette data is poked directly to the RAMDAC.

-- Input:
-- Color    :   0 ==> 255   :   1-byte
-- Values   :   {R, G, B}   :   3-bytes

global procedure SetPalette(integer Color, sequence Values)
    
    poke(SetVGAPalette,
    {#60,                   --    0: pushad
    #66,#BA,#C8,#03,        --    1: mov dx, #03C8
    #B0,Color,              --    5: mov al, color (6)
    #EE,                    --    7: out dx, al
    #66,#FF,#C2,            --    8: inc dx
    #B0,Values[1],          --    B: mov al, red (12)
    #EE,                    --    D: out dx, al
    #B0,Values[2],          --    E: mov al, green (15)
    #EE,                    --   10: out dx, al
    #B0,Values[3],          --   11: mov al, blue (18)
    #EE,                    --   13: out dx, al
    #61,                    --   14: popad
    #C3}                    --   15: ret
    )
    call(SetVGAPalette)
    
end procedure
integer PokePalette PokePalette = routine_id("SetPalette")

--------------------------------------------------------------------------------

-- sequence = GetPalette(integer color)
-- Returns the desired color palette as a sequence of RGB palette data.

-- Input:
-- Color    :   0 ==> 255   :   1-byte

-- Output:
-- Values   :   {R, G, B}   :   3-bytes

global function GetPalette(integer Color)
    
    poke(GetVGAPaletteCol, Color)
    call(GetVGAPalette)
    
    return peek({stack, 3})

end function

--------------------------------------------------------------------------------

-- sequence = PhongBlend(sequence palarray,
--      integer Color1, integer Color2, sequence palette1, sequence palette2)

-- Creates a smooth blend from color to color useing palette to palette.
-- Great on effects such as fire, water, plasma etc...

-- Input:
-- Palette  :   {{R, G, B}, {R, G, B}, etc...}  :   256 x 3-bytes
-- Pal1     :   0 ==> 255                       :   1-byte
-- Pal2     :   0 ==> 255                       :   1-byte
-- RGB1     :   {R, G, B}                       :   3-bytes
-- RGB2     :   {R, G, B}                       :   3-bytes

-- Output:
-- Palette  :   {{R, G, B}, {R, G, B}, etc...}  :   256 x 3-bytes

global function PhongBlend(sequence Palette,
		    integer Pal1, integer Pal2, sequence RGB1, sequence RGB2)

    integer Point
    atom Red, Green, Blue
    sequence RGB
    
    RGB = repeat(0, 3)
    Red     =   RGB2[1] - RGB1[1]
    Green   =   RGB2[2] - RGB1[2]
    Blue    =   RGB2[3] - RGB1[3]
    Point   =   Pal2 - Pal1
    for i = 0 to Point do
	RGB[1] = floor(Red / Point * i)     + RGB1[1]
	RGB[2] = floor(Green / Point * i)   + RGB1[2]
	RGB[3] = floor(Blue / Point * i)    + RGB1[3]
	Palette[Pal1 + (i + 1)] = RGB
    end for
    
    return Palette
    
end function

--------------------------------------------------------------------------------

-- PhongIllumination(sequence Palette,
--          integer Pal1, integer Pal2, sequence R, sequence G, sequence B)

-- Mimics natural behaviour of light from color to color given ambient,
-- diffuse, specular and light values for red, green and blue.

-- Input:
-- Palette  :   {{R, G, B}, {R, G, B}, etc...}  :   256 x 3-bytes
-- Pal1     :   0 ==> 255                       :   1-byte
-- Pal2     :   0 ==> 255                       :   1-byte

-- Each R, G & B are a four element sequence wich contain the following;

-- Element[1]   =   Ambient     :   All-present light   :   1-byte
-- Element[2]   =   Diffuse     :   Gradual rise        :   1-byte
-- Element[3]   =   Specular    :   Reflection          :   1-byte
-- Element[4]   =   light       :   Intensity           :   1-byte

-- Output:
-- Palette  :   {{R, G, B}, {R, G, B}, etc...}  :   256 x 3-bytes

-- * n.b While the higher palettes of the selected palette field seem to become
-- white they are actualy natural light shades. If you interpolate, blur, wrap
-- or apply any other effect on theese shades you will see that they will
-- produce some nice colorfull effects.

global function PhongIllumination(sequence Palette,
	    integer Pal1, integer Pal2, sequence R, sequence G, sequence B)

    atom Angle, Step
    sequence RGB
    
    RGB     = repeat(1, 3)
    Angle   = PI / 2
    Step    = Angle / (Pal2 - Pal1)
    for i = 0 to Pal2 - Pal1 do
	
	for j = 1 to R[4] do
	    RGB[1] = (RGB[1] * R[3] * cos(Angle)) + (R[1] + R[2] * cos(Angle))
	    if RGB[1] > 63 then
		RGB[1] = 63
	    end if
	end for
	
	for j = 1 to G[4] do
	    RGB[2] = (RGB[2] * G[3] * cos(Angle)) + (G[1] + G[2] * cos(Angle))
	    if RGB[2] > 63 then
		RGB[2] = 63
	    end if
	end for
    
	for j = 1 to B[4] do
	    RGB[3] = (RGB[3] * B[3] * cos(Angle)) + (B[1] + B[2] * cos(Angle))
	    if RGB[3] > 63 then
		RGB[3] = 63
	    end if
	end for

    Angle -= Step
    Palette[Pal1 + (i + 1)] = floor(RGB)
    end for
    
    return Palette
    
end function

--------------------------------------------------------------------------------

-- sequence = PhongShade(sequence Palette,
--                              integer Red, integer Green, integer Blue)

-- Creates smooth balck and white palette entries given a valid palette array.
-- Recommended Intensity is; Red: 30% Green: 60% Blue: 15% for best results.

-- Input:
-- Palette  :   {{R, G, B}, {R, G, B}, etc...}  :   256 x 3-bytes
-- Red      :   1 ==> 100                       :   1-byte
-- Green    :   1 ==> 100                       :   1-byte
-- Blue     :   1 ==> 100                       :   1-byte

-- Output:
-- Palette  :   {{R, G, B}, {R, G, B}, etc...}  :   256 x 3-bytes

global function PhongShade(sequence Palette,
				integer Red, integer Green, integer Blue)
    
    for i = 1 to 256 do
	Palette[i] = repeat(floor(((Palette[i][1] * Red) /  100)
			    +   ((Palette[i][2] * Green) /  100)
			    +   ((Palette[i][3] * Blue) /   100)),
			    3)
    end for
    
    return Palette
    
end function

--------------------------------------------------------------------------------

constant VertRetrace = allocate(18)     --> waits until vertical retrace
poke(VertRetrace, {
    #60,                    --    0: pushad
    #66,#C7,#C2,#DA,#03,    --    1: mov dx, #03DA
    #EC,                    --    6: @wait1: in al, dx
    #A8,#08,                --    7: test al, #08
    #75,#FB,                --    9: jnz @wait1
    #EC,                    --    B: @wait2: in al, dx
    #A8,#08,                --    C: test al, #08
    #74,#FB,                --    E: jz @wait2
    #61,                    --   10: popad
    #C3})                   --   11: ret

--------------------------------------------------------------------------------

-- BlackOut()

-- Blacks out the screen setting all RGB palette values to zero. This clears
-- the screen without destroying the screens contence. Mostly usefull when
-- changing between different palette arrays.

global procedure BlackOut()
    sequence Palette
    
    Palette = repeat({0, 0, 0}, 256)
    SetAllPalette(Palette)
    
end procedure

--------------------------------------------------------------------------------

-- FadeIn(sequence Palette, integer Seconds)

-- Smoothly fades the current palette table to the desired pallette table. All
-- palette values will reach thier desired destination at the same time.
-- Seconds determines approximatly how long the fade process should take.

-- Input:
-- Palette  :   {{R, G, B}, {R, G, B}, etc...}  :   256 x 3-bytes
-- Seconds  :   0 ==> +                         :   float

global procedure FadeIn(sequence Palette, atom Seconds)
    sequence PaletteMask, RGB
    atom x, y, z, s
    
    RGB = repeat(0, 3)
    Seconds *= 70
    s = 1
    while Seconds > 327 do
	Seconds /= 2
	s *= 2
    end while
    PaletteMask = call_func(PeekAllPalettes, {})
    
    x = Seconds * 256
    for i = 1 to Seconds do
	y = i * 256
	for j = 1 to 256 do
	    z = (((Palette[j][1] - PaletteMask[j][1]) * 256) * y) / x
	    RGB[1] = (PaletteMask[j][1] + z) / 256
	    z = (((Palette[j][2] - PaletteMask[j][2]) * 256) * y) / x
	    RGB[2] = (PaletteMask[j][2] + z) / 256
	    z = (((Palette[j][3] - PaletteMask[j][3]) * 256) * y) / x
	    RGB[3] = (PaletteMask[j][3] + z) / 256
	    call_proc(PokePalette, {j - 1, RGB})
	end for
	
	for j = 1 to s do
	    call(VertRetrace)
	end for
    end for
    call(VertRetrace)
    call_proc(PokeAllPalettes, {Palette})
    
end procedure

--------------------------------------------------------------------------------

-- FadeOut(integer Seconds)

-- Smoothly fades out the screen at the rate of approximate seconds. All
-- palette values will reach zero at the same time.

-- Input:
-- Seconds  :   0 ==> + :   float

global procedure FadeOut(atom Seconds)
    sequence Palette, RGB
    atom x, y, z, s
    
    RGB = repeat(0, 3)
    Seconds *= 70
    s = 1
    while Seconds > 127 do
	 Seconds /= 2
	 s *= 2
    end while
    Palette = call_func(PeekAllPalettes, {})

    x = Seconds * 256
    for i = 1 to Seconds do
	y = i * 256
	for j = 1 to 256 do
	    z = ((Palette[j][1] * 256) * y) / x
	    RGB[1] = Palette[j][1] - z / 256
	    z = ((Palette[j][2] * 256) * y) / x
	    RGB[2] = Palette[j][2] - z / 256
	    z = ((Palette[j][3] * 256) * y) / x
	    RGB[3] = Palette[j][3] - z / 256
	    call_proc(PokePalette, {j - 1, RGB})
	end for

	for j = 1 to s do
	    call(VertRetrace)
	end for
    end for

end procedure

--------------------------------------------------------------------------------

-- sequence = RotatePalette(sequence Palette, integer Color1, integer Color2)

-- Rotates the palatte either clockwise or counter-clockwise.

-- Input:
-- Palette  :   {{R, G, B}, {R, G, B}, etc...}  :   256 x 3-bytes
-- Pal1     :   0 ==> 255                       :   1-byte
-- Pal2     :   0 ==> 255                       :   1-byte

-- Output:
-- Palette  :   {{R, G, B}, {R, G, B}, etc...}  :   256 x 3-bytes

global function RotatePalette(sequence Palette, integer Pal1, integer Pal2)
    
    Pal1 += 1
    Pal2 += 1
    
    if Pal2 > Pal1 then
	for i = 1  to Pal2 - Pal1 do
	    Palette = {Palette[256]} & Palette[1..255]
	end for
	
    elsif Pal1 > Pal2 then
	for i = 1 to  Pal1 - Pal2 do
	    Palette = Palette[2..256] & {Palette[1]}
	end for
    end if

    return Palette
    
end function

--------------------------------------------------------------------------------
--============================================================================--
